unit MpegStream;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls,BitStream,Math,Mpegutils;

  Type
   TData=array[0..2-1] of integer;


   Type
   TPictures = class
    private
      // The data fields of this new class
    public
      // Properties to read these data values
     FrameBuffer:array of array of Integer;
    // Constructor
      constructor Create(FWidth,FHeight:Integer);
      Destructor Destroy; override;
  end;


  Type
  TMpegStream = class
    private
     Width :Integer;
     Height:Integer;
     Block: array[0..6-1, 0..64-1] of Smallint;
     quantization_scale:integer;
     idc_dct_pred:array[0..3-1]of integer;
     vblock:integer;
     hblock:integer;
     Y_block:array[0..256-1] of  Smallint;         //16x16 ( four 8x8 samples )
     Cb_block:array[0..64-1] of  Smallint; 		//8x8
     Cr_block:array[0..64-1] of  Smallint; 		//8x8
     Stream:TBitsStream;
      procedure Reset_dc_dct_pred();
      procedure DecodeBlock(Pictures:TPictures);
      function GetIntraDCLumValue():integer;
      function GetIntraDCChromValue():integer;
      function GetACValue():TData;
      procedure GetIntraBlock(var block:array of Smallint;cc:integer);
      procedure GetMacroblock(Pictures:TPictures);
      procedure GetSlice(Pictures:TPictures; code:integer);
      procedure GetPictureFrame();
      procedure GetSequenceHeader();
      procedure GetGroupPictures();
      Function DecodeMpeg():Integer;

    public
     PictureList : TList;
     Pictures :TPictures;
      constructor Create(FileName:String);
      Destructor Destroy; override;
      procedure Drawbitmap(FrameNo:Integer;Image:TImage);
  end;


implementation

Uses Unit1;

constructor TPictures.Create(FWidth,FHeight:Integer);
begin
  SetLength( FrameBuffer,FWidth ,FHeight);
end;

destructor TPictures.Destroy;
begin
  inherited;
end;




constructor TMpegStream.Create(FileName:String);
var
I:Integer;
begin
  Stream:=TBitsStream.Create(FileName);
  PictureList := TList.Create;

  For I:=0 To 3-1 do
  idc_dct_pred[I]:=128;
  vblock:=0;
  hblock:=0;

 DecodeMpeg();
end;

destructor TMpegStream.Destroy;
begin
  Stream.Free;
  PictureList.Free;
  inherited;
end;

procedure TMpegStream.Reset_dc_dct_pred();
begin
  idc_dct_pred[0] := 128;
  idc_dct_pred[1] := 128;
  idc_dct_pred[2] := 128;
end;


procedure TMpegStream.Drawbitmap(FrameNo:Integer;Image:TImage);
VAR
  Bitmap:  TBitmap;
  i,j   :  INTEGER;
  Pictures:TPictures;
begin
   Pictures:=TPictures(PictureList[FrameNo]);
   Bitmap := TBitmap.Create;
   Try
    Bitmap.PixelFormat := pf24bit;
    Bitmap.Width  := Width;
    Bitmap.Height := Height;
  For j := 0 TO Height-1 Do
    For i := 0 TO Width-1 Do   Begin
     Bitmap.Canvas.Pixels[i,j]:=Pictures.FrameBuffer[i,j];
    End;
    Image.Picture.Graphic := Bitmap;
  Finally
  Bitmap.Free;
  End;
end;

procedure TMpegStream.DecodeBlock(Pictures:TPictures);
var
 i,j,m,k,offset:integer;
 Y,CB,CR,R,G,B:Byte;
 index :integer;
begin
  offset :=0;
  index := 0;

 	for  i := 0 to 8-1 do begin
			for  j :=0 to   8-1  do begin
				Y_block[offset] := block[0][index];
				Y_block[offset + 8] := block[1][index];
				Y_block[offset + 128] := block[2][index];
				Y_block[offset + 8 + 128] := block[3][index];
        inc(offset);
				inc(index);
			end;
		 inc(	offset, 8);
	end;

    for  j :=0 to   64-1  do begin
       Cb_block[j]:=Block[4,j];
       Cr_block[j]:=Block[5,j];
    end;



   if hblock>(Width-1) div 16 then
   begin
    hblock:=0;
    inc(vblock);
  end;

  m :=0;
  k := 0;
 for i:=0 to 16-1 do
   for J:=0 to 16-1 do
   begin
      if  (not(boolean( i and 1))) and (not(boolean( j and 1))) then
      begin    //one Cb, Cr has been saved for every 4 pixels
        Y  := Y_block[m];
        Cb := Cb_block[k] ;
        Cr := Cr_block[k] ;
        YCbCrToRGB(Y, Cb, Cr, R, G, B);
        inc(k);
      end else
      begin
         Y := Y_block[m];
         YCbCrToRGB(Y, Cb, Cr, R, G, B);
      end;
      inc(m);
      Pictures.FrameBuffer[(hblock*16)+j, (vblock*16)+i]:=RGB(R,G,B);
      end;
   inc(hblock);
end;

function TMpegStream.GetIntraDCLumValue():integer;
var
 code,nbits, dct_diff:integer;
begin

    code := Stream.showBits(9);
		if (code >= 504)then
			dec(code ,488)
		else if (code >= 448) then
			code := (code shr 3) - 48
		else
			code :=code shr 6;
	 	Stream.flushBits(DClumtable[code][1]);
	  nbits := DClumtable[code][0];

    if  (nbits >0) then  begin
       code := Stream.getBits(nbits);
       if ((code and (1 shl (nbits - 1))) <> 0) then  begin
         dct_diff:=code
       end
       else begin
        dct_diff:=((-1 shl nbits) or (code+1));
       end;
    end
    else begin
    dct_diff:=0;
    end;

  result:=dct_diff;

end;

function TMpegStream.GetIntraDCChromValue():integer;
var
 code,nbits, dct_diff:integer;
begin

		code := Stream.showBits(10);
		if (code >= 1016) then
			code:=code-992
		else if (code >= 960) then
			code := (code shr 3) - 104
		else
			code :=code shr 6;
	  	Stream.flushBits(DCchrtable[code][1]);
      nbits := DCchrtable[code][0];

	 if  (nbits >0) then  begin
       code := Stream.getBits(nbits);
       if ((code and (1 shl (nbits - 1))) <> 0) then  begin
         dct_diff:=code
       end
       else begin
        dct_diff:=((-1 shl nbits) or (code+1));
       end;
    end
    else begin
    dct_diff:= 0;
    end;

  result:=dct_diff;
end;

function TMpegStream.GetACValue():TData;
var
  code:integer;
  Data:TData;
begin

	 code := Stream.showBits(16);
		if (code >= 10240) then
			code := (code shr 11) + 112
		else if (code >= 8192) then
			code := (code shr 8) + 72
		else if (code >= 1024) then
			code := (code shr 9) + 88
		else if (code >= 512) then
			code := (code shr 6) + 72
		else if (code >= 256) then
			code := (code shr 4) + 48
		else if (code >= 128) then
			code := (code shr 3) + 32
		else if (code >= 64)  then
			code := (code shr 2) + 16
		else if (code >= 32)  then
			code :=code shr 1
		else if (code >= 16) then
			dec(code ,16)
     else exit;


		Stream.flushBits(DCTtable[code][1]);

		Data[0] := DCTtable[code][0] and $FF;
		Data[1] := DCTtable[code][0] shr 8;
		if (data[0] = ESC) then begin
			Data[0] := Stream.getBits(6);
			Data[1] := Stream.getBits(8);
			if (Result[1] = $00) then
				Data[1] := Stream.getBits(8)
			else if (data[1] = $80) then
				Data[1] := Stream.getBits(8) - 256
			else if (data[1] >= $80) then
				dec(data[1] , 256);
		end else if (data[0] <> EOB)then begin
			if (Stream.getBits(1) <> 0) then
				data[1] := -data[1];
		end;
   result:=data;
end;


procedure TMpegStream.GetIntraBlock(var block:array of Smallint;cc:integer);
var
i,j:integer;
data:TData;
position:integer;
dct_diff:Integer;
level:integer;
begin

   if cc=0 then
    idc_dct_pred[cc]:=idc_dct_pred[cc]+ GetIntraDCLumValue()
   else
    idc_dct_pred[cc]:=idc_dct_pred[cc]+ GetIntraDCChromValue() ;

    block[0]:= idc_dct_pred[cc] shl 3 ;

  i:=1;
  level:=0;
  while i<64-1 do
  begin
      data:= getACValue();
			if (data[0] =EOB) then  break;
         i := (i+ data[0])and 63;
         position := zigzag[i];
         block[position]:= data[1]*quantization_scale*intraMatrix[position]  shr 3 ;
       inc(i);
  end;
end;


procedure TMpegStream.GetMacroblock(Pictures:TPictures);
var
i,j:integer;
cc:integer;
begin
  for i:=0 to 6-1 do
  begin
    for j:=0 to 64-1 do block[i,j]:=0;
    if i<4 then cc:=0  else cc:=i-3;
      GetIntraBlock(block[i],cc);
      idct(block[i]);
  end;
  stream.getBits(2);
  DecodeBlock(Pictures);
end;



procedure TMpegStream.GetSlice(Pictures:TPictures; code:integer);
begin

 quantization_scale:=stream.getBits(5);
 stream.getBits(1);
 Stream.alignBits(8);

 while (stream.showBits(23) <> 0) do
  begin
    GetMacroblock(Pictures);
  end;

end;


procedure TMpegStream.GetPictureFrame();
var
  temporal_reference  :Integer;
  picture_coding_type :Integer;
  vbv_delay          :Integer;
  Pictures:TPictures;
begin
  vblock:=0;
  hblock:=0;

  Reset_dc_dct_pred();

  temporal_reference  := Stream.GetBits(10);
  picture_coding_type := Stream.GetBits(3);
  vbv_delay           := Stream.GetBits(16);



  if (picture_coding_type = I_TYPE) then
  begin
     Pictures:=TPictures.Create(Width*3,Height*3);

      while (stream.getBits(1) <> 0) do
	 		stream.flushBits(8);

      while (stream.showCode() = EXT_START_CODE) or (stream.showCode() = USER_START_CODE) do
      stream.getCode();

      while (stream.showCode() >= SLICE_MIN_CODE)and (stream.showCode() <= SLICE_MAX_CODE) do
      begin
        GetSlice(Pictures,stream.getCode());
        Reset_dc_dct_pred();
      end;

     PictureList.Add(Pictures);
   end;

  
end;

procedure TMpegStream.GetSequenceHeader();
begin
   vblock:=0;
   hblock:=0;
   Width:=Stream.getbits(12);
   Height:=Stream.getbits(12);
end;

procedure TMpegStream.GetGroupPictures();
begin

end;


Function TMpegStream.DecodeMpeg():Integer;
begin
  while (stream.showCode() <> SEQ_END_CODE) do
  begin
			case (stream.getCode()) of
			  SEQ_START_CODE: GetSequenceHeader();
        GOP_START_CODE: GetGroupPictures();
        PIC_START_CODE: GetPictureFrame();
      end;
  end;
end;


end.
